#ifndef PAR_H
#define PAR_H

// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of paramxx
//
// paramxx is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// paramxx is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <iostream>
#include <string>
#include <cstring>
#include <vector>

using namespace std;

#include "ParFileException.h"
#include "ParTxt.h"

/**
 * The base class for the boolean, double, integer and string parameters.
 */
class Par {

  /**
   * This operator outputs the contents of the object.
   */
  friend ostream& operator << ( ostream& os, Par& par ) {
    par.print( os );
    return os;
  }

  /**
   * This operator outputs the contents of the object.
   */
  friend ostream& operator << ( ostream& os, Par* par ) {
    par->print( os );
    return os;
  }
  
public:

  /**
   * The delimit within the par file
   */
  enum ParFileDelimit { DELIMIT=',' };

  /**
   * The position of the parameters.
   */
  enum ParType { PARNAME,    /**< The name of the parameter */
		 PARTYPE,    /**< The type of the parameter */
		 PARMODE,    /**< The mode of the parameter */
		 PARVALUE,   /**< The value of the parameter */
		 PARMINIMUM, /**< The minimum value of the parameter */
		 PARMAXIMUM, /**< The maximum value of the parameter */
		 PARPROMPT   /**< The name of the parameter */
  };

  enum ParamMode { AUTOS='a', BATCH='b', hIDDEN='h',
		   HIDDEN='H', QUERRY='q', LEARN='l' };

  enum ParamType { BOOLEAN='b', COMMENT='c', INTEGER='i', 
		   REAL='r',  STRING='s' };

  enum NumTokens { NUMTOKENS=7 };

  /**
   * The destructor.
   */
  virtual ~Par( void ) { };

  /**
   * The default constructor.
   */
  Par( void ) { }

  Par( ParTxt& par );

  int check_value( const char* str ) const;

  /**
   * All the derived Par class, with the exception of the comment class,  will 
   * call this function to print on the screen the parameters with the format
   *                        name = val   prompt
   */
  virtual void plist( ostream& os ) const;

  /**
   * Get the name member of the class.
   */
  string get_name( void ) const { return parameter[ PARNAME ]; }

  /**
   * Get the mode of the parameter file	
   */
  string get_mode( void ) const { return parameter[ PARMODE ]; }

  /**
   * Get the prompt of the parameter
   */
  string get_prompt( void ) const { return parameter[ PARPROMPT ]; }
  
  string get_value( void ) const { return parameter[ PARVALUE ]; }
  
  static int my_tokenize( char* str, char* delimit, char*** tokens )
    throw ( Exception );
  static void delete_tokens( char** ptr );

  /**
   * The user is trying to get the boolean value of
   * a parameter which is anything but a boolean.
   */
  virtual bool pgetb( void ) const throw ( ParFileException );
  /// only defined in BoolPar

  /**
   * The user is trying to get the double value of
   * a parameter which is anything but a double.
   */
  virtual double pgetd( void ) const throw ( ParFileException );
  /// only defined in RealPar

  virtual int pgeti( void ) const throw ( ParFileException );
  /// only defined in LongPar
  
  virtual long pgetl( void ) const throw ( ParFileException );
  /// only defined in LongPar
  
  virtual void pgetstr( char result[], size_t size ) const
	throw ( ParFileException );
  /// only defined in StringPar

  virtual string pgetstring( void ) const throw ( ParFileException );
  /// only defined in StringPar

  /**
   * To output the parameter of the form: name,type,mode,val,min,max,prompt
   */
  virtual void print( ostream& os ) const;

  virtual void set_val( const string& str )
    throw ( ParFileException, Exception );

  static bool is_indirrect( const string& str, char delimit=')' );

protected:

  vector< string > parameter;

  void not_between_limits( char str[], 
			   const char* left, const string& left_val,
			   const char* right, const string& right_val ) const;

private:

  static int count_tokens( char* str, char* delimit ) throw ( Exception );
  static void print_tokens( char* str, char** str_argv, ostream& os );
  static char* save_string( char* str ) throw ( Exception );

  /**
   * To see if the value is between the lower and upper
   * limit of the parameter defined in all derived classes
   */
  virtual void between_limits( const string& str ) const
    throw ( ParFileException, Exception );

  /**
   * To be called by pgetb, pgetd, pgeti, pgetl, pgetstr, pgetstring if
   * the type requested do not match the type of parameter.
   */
  void pget_wrong_type( char[], const char* ) const;

  virtual void validate( const string& ) const throw ( ParFileException );
  /// to be defined in all derived classes

};

#endif
