// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of paramxx
//
// paramxx is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// paramxx is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <unistd.h>
#include <fcntl.h>
#include <fstream>
#include <cstdlib>
#include <cstring>

using namespace std;

#include <suplibxx/str.h>

#include "ParFilename.h"

// The order which the environment variables shall be search for:
static const char* environment_variables[] = { "PDIRS", "PFILES", "UPARM" };
#define NUM_ENV_VAR sizeof(environment_variables)/sizeof(char *)

// The order which the file extension shall be searched for:
static const char* pfile_extension[] = { ".par" };
#define NUM_EXTENSION sizeof(pfile_extension)/sizeof(char *)

typedef vector< string > myVector;
typedef myVector::const_iterator myVectorIterator;

ParFilename::ParFilename( int argc, char** argv, const char* file ) 
  throw ( ParFileException ) {

  try {
    init_ParFilename( argc, argv, file );
  } catch( ParFileException& pfe ) {
    throw;
  }

}

int ParFilename::accessible( const char* str ) throw ( ParFileException ) {

  try {
    // if the filename is an empty string 
    // then obviously the file is not accessible.
    if ( 0 == str || 0 == *str ) 
      return EXIT_FAILURE;

    if ( 0 == ::access( str, R_OK ) ) {
      if ( '/' == str[ 0 ] ) {
	// copy file from wherever to local working directory
	// comment out for now, will implement using the correct requirements
	// copy_pfile( str );
	// Since we no longe copy the file to the current directory
	// filename = "./";
	// filename += get_basename( str );
      }
      filename = str;

      return EXIT_SUCCESS;
    }

  } catch( ParFileException& pfe ) {

    string msg( "ParFilename::accessible( " );
    msg.append( str );
    msg.append( " )\n" );
    pfe.set_message( msg );
    throw pfe;

  }
   
  return EXIT_FAILURE;

}

void ParFilename::check_environment_variables( ) throw ( ParFileException ) {

  // Make sure that at least one of the environment variables is set.
  // Keep a counter of the number of environment variables which is not set.
  int env_var_counter = 0;
  {
    for ( size_t i = 0; i < NUM_ENV_VAR; i++ )
      if ( NULL == getenv( environment_variables[ i ] ) )
	env_var_counter++;
  }

  if ( NUM_ENV_VAR == env_var_counter ) {

    // If the number of env_var_counter is equal to the number of environment 
    // variables then not a single environment variable is set.
    string msg = "ParFilename::check_environment_variables( ) : "
      "One of the following environment variables ";
    
    for ( size_t i = 0; i < NUM_ENV_VAR; i++ ) {
      msg.append( "`" );
      msg.append( environment_variables[ i ] );
      msg.append( "'" );
      if ( i != NUM_ENV_VAR - 1 )
	msg.append( ", " );
    }
     
    msg.append( " must be defined\n" );

    throw ParFileException( msg );

  }

  return;

}

void ParFilename::copy_pfile( const char* str ) 
  const throw ( ParFileException ) {

  int f1, f2, n;
  char buf[BUFSIZ];

  if ( -1 == (f1 = open( str, O_RDONLY, 0 ) ) ) {
    string msg( "ParFilename::copy_pfile( " );
    msg.append( str );
    msg.append( " ) : Unable to open file for reading\n" );
    throw ParFileException( msg );
  }

  if ( -1 == (f2 = creat( get_basename( str ), 0644 ) ) ) {
    string msg( "ParFilename::copy_pfile( " );
    msg.append( str );
    msg.append( ", ostream& ) : Unable to open file for writing\n" );
    close( f1 );
    throw ParFileException( msg );
  }

  while( (n = read( f1, buf, BUFSIZ ) ) > 0 )
    if ( n != write( f2, buf, n ) ) {
      string msg( "ParFilename::copy_pfile( " );
      msg.append( str );
      msg.append( " ) : Error writing to file" );
      close( f1 );
      close( f2 );
      throw ParFileException( msg );
    }

  close( f1 );
  close( f2 );

  return;
  
}

int ParFilename::find_pfile( const char* name ) throw ( ParFileException ) {

  try {

    // The absolute pathname was given with the correct extension.
    // Make sure that it contains an extention otherwise the library
    // may have mistaken the executable for the parfile.
    if ( strstr( name, "." ) && EXIT_SUCCESS == accessible( name ) )
      return EXIT_SUCCESS;

    for ( size_t ii = 0; ii < NUM_ENV_VAR; ii++ ) {

      char* env_var = getenv( environment_variables[ ii ] );

      if ( env_var ) {

	string directories( env_var );

	vector< string > directory;
	suplib::tok( directory, directories, ":;" );

	/*
	vector< string > directory =
	  parse_string( s, pfiles_delimit );
	*/

	for ( myVectorIterator iter = directory.begin( );
	      iter != directory.end( ); ++iter ) {

	  char fname[ 1024 ];
	  sprintf( fname, "%s/%s.par", iter->data( ), name );

	  if( EXIT_SUCCESS == accessible( fname ) )
	    return EXIT_SUCCESS;

	}

      }

    }

    return EXIT_FAILURE;

  } catch ( ParFileException& pfe ) {

    throw;

  }

}

const char* ParFilename::get_basename( const char* name ) const {

  if ( name || *name ) {

    const char* progname = name;

    char dirdelimit[ 2 ] = { dir_delimit, '\0' };

    const char* ptr = name;

    do {

      ptr = strstr( progname, dirdelimit );
      if ( ptr )
	progname = ptr + 1;

    } while( ptr );

    return progname;

  } else
    return NULL;

}

void ParFilename::init_ParFilename( int argc, char** argv,
				    const char* file ) 
  throw ( ParFileException ) {

  try {

    check_environment_variables( );

    // loop thru the cmdline arguments.  If PFILE is one of the
    // options then check if the requested file is present
    for ( int i = 1; i < argc; i++ ) {

      const char* ptr = strstr( argv[ i ], PFILE );
      if ( ptr ) {
      
	ptr += strlen( PFILE );
	if ( EXIT_SUCCESS == find_pfile( ptr ) )
	  return;

      }

    }

    //
    // If a specific file was requested then it takes
    //  preference over the default file ( using argv[0] ).
    //
    if ( file ) {
      if ( EXIT_SUCCESS == find_pfile( file ) )
	return;

      string msg( "ParFilename::init_ParFilename( ) : "
		  "Unable to find parfile `" );
      msg.append( file );
      msg.append( "'\n" );

      throw ParFileException( msg ) ;

    }

    // argv[ 0 ] contains the full pathname.  Therefore,
    // one must find the basename of the executable.
    const char* progname = get_basename( argv[ 0 ] );

    if ( EXIT_SUCCESS == find_pfile( progname ) )
      return;

  } catch ( ParFileException& pfe ) {

    throw pfe;

  }

  string msg( "ParFilename::init_ParFilename( ) : "
	      "Unable to find parfile for `" );
  msg.append( argv[ 0 ] );
  msg.append( "'\n" );

  throw ParFileException( msg );

}

/**
 * \param str the string to be split.
 * \param delimiter the delimit to split the string.
 *
 * Splits the input line on the parameter delimiter.
 *
 * \return Vector of tokens, one for each delimiter delimited field.
 */
vector<string> ParFilename::parse_string( string& str, char delimit ) {

  vector<string> tokens;
 
  if ( str.size( ) ) {

    size_t st0 = 0, st1 = 0;
    do {

      st1 = str.find( delimit, st0 );
      if ( str.size( ) < st1 ) 
	st1 = str.size( );

      tokens.push_back( str.substr( st0, st1 - st0 ) );

      st0 = st1 + 1;

    } while ( str.size( ) > st1 );

  }

  return tokens;
}

#undef NUM_ENV_VAR
#undef NUM_EXTENSION
